/*******************************************************************************
 * (c) Copyright 2016-2017 Microsemi SoC Products Group. All rights reserved.
 * 
 * This SoftConsole Video project for MIPI sensor configuration and interfacing with GUI
 *
 * Please refer README.TXT in the root folder of this project for more details.
 */
#include "riscv_hal.h"
#include "hw_platform.h"
#include "core_uart_apb.h"
#include "core_gpio.h"
#include "core_timer.h"
#include "core_i2c.h"


/* Fabric Addresses */
#define Brightness_Gain_Address  	0x70009600
#define Contrast_Gain_Address   	0x70009604
#define Saturation_Gain_Address 	0x70009608
#define Hue_Gain_Address        	0x7000960C
#define Sharpness_K_factor      	0x70009610
#define AB_Image_Height         	0x70009614
#define AB_Image_Width          	0x70009618
#define AB_Image_PosX           	0x7000961C
#define AB_Image_PosY           	0x70009620
#define AB_Image_GB_Alpha       	0x70009624
#define AB_Image_Address        	0x70009628
#define Output_scalar_resolution	0x7000962C

#define SIN_MEM_BASE_ADDR        	0x70009300
#define COS_MEM_BASE_ADDR        	0x70009000

const char * g_hello_msg =
"\r\n         MIV  video configuration        \n\r\n\r";

volatile uint32_t g_10ms_count;

volatile uint32_t timerdone = 0;
volatile uint32_t g_10ms_count1;
volatile uint32_t g_ms_count;

i2c_instance_t g_i2c_instance;
i2c_instance_t g_i2c_instance1;

/*-----------------------------------------------------------------------------
 * UART instance data.
 */
UART_instance_t g_uart;
  
/*-----------------------------------------------------------------------------
 * GPIO instance data.
 */

gpio_instance_t g_gpio_in;
gpio_instance_t g_gpio_out;

/*-----------------------------------------------------------------------------
 * Global state counter.
 */
uint32_t g_state = 1;

volatile uint32_t rx_tmr_done = 0;
volatile uint32_t rx_ms_count1;
volatile uint32_t rx_ms_count;
uint32_t t_ms_count = 0;


/*-----------------------------------------------------------------------------
 * UART handler specific.
 */
#define MAX_RX_DATA_SIZE    1024
#define MAX_TX_DATA_SIZE    1024
uint8_t g_rx_data[MAX_RX_DATA_SIZE];
uint8_t g_tx_data[MAX_TX_DATA_SIZE] = {'a'};
uint8_t g_rx_data1[MAX_RX_DATA_SIZE];
uint8_t rx_byte;
volatile uint32_t rx_cnt = 0;
uint32_t i = 0;
uint32_t rx_size;
uint32_t process_data = 0;

void UART_Handler();

uint16_t generateCRC(const uint8_t * data, int length);

uint32_t packet_no = 0;

volatile uint32_t *sin_mem_base_addr = (volatile uint32_t *) SIN_MEM_BASE_ADDR;
volatile uint32_t *cos_mem_base_addr = (volatile uint32_t *) COS_MEM_BASE_ADDR;

/*-----------------------------------------------------------------------------
 * System Tick interrupt handler
 */
void SysTick_Handler(void) {

	g_state = (~g_state) & 0x01;

	GPIO_set_output(&g_gpio_out, GPIO_2, g_state);
	if(timerdone == 1)
	{
		g_10ms_count1 += 1;
		if(g_ms_count <= g_10ms_count1)
			timerdone = 0;
	}

	if(rx_tmr_done == 1)
	{
		rx_ms_count1 += 1;
		if(rx_ms_count1 >= rx_ms_count){
			rx_tmr_done = 0;
			process_data = 1;
		}
	}
}

void External_29_IRQHandler(void) {

	I2C_isr(&g_i2c_instance1);
}

void External_28_IRQHandler(void) {

	I2C_isr(&g_i2c_instance);
}

/*-----------------------------------------------------------------------------
 * main
 */
int main(int argc, char **argv) {
    uint8_t rx_char;
    uint8_t rx_count;
    volatile  uint32_t i, temp;
    unsigned long j;
	uint32_t  key;
	uint16_t crc16;
	uint16_t rcvd_crc16;


    PLIC_init();

    GPIO_init(&g_gpio_in, COREGPIO_IN_BASE_ADDR, GPIO_APB_32_BITS_BUS);
    GPIO_init(&g_gpio_out, COREGPIO_OUT_BASE_ADDR, GPIO_APB_32_BITS_BUS);

    UART_init(&g_uart,
              COREUARTAPB0_BASE_ADDR,
              BAUD_VALUE_115200,
              (DATA_8_BITS | NO_PARITY));

    printf("Microsemi Video IP Configuration\n");

    temp = read_csr(mstatus);

    SysTick_Config(SYS_CLK_FREQ / 1000);

    PLIC_SetPriority(I2C_IRQn, 1);

    PLIC_EnableIRQ(I2C_IRQn);

    PLIC_SetPriority(HDMI_I2C_IRQn, 1);

    PLIC_EnableIRQ(HDMI_I2C_IRQn);

    HAL_enable_interrupts();

	GPIO_set_output(&g_gpio_out, GPIO_8, 0u);
	for(i=0; i < 10000; i++);
	GPIO_set_output(&g_gpio_out, GPIO_8, 1u); // Bring camera out of reset


	mipi_1080_720_30fps_10bit_49mp_4lane();
	HDMI_tx_init();

	sincos_mem_init();

	*(volatile int*) Brightness_Gain_Address = 0x0000;
	*(volatile int*) Contrast_Gain_Address = 0x0000;
	*(volatile int*) Saturation_Gain_Address = 0x001;
	*(volatile int*) Hue_Gain_Address = 0x0000;
	*(volatile int*) Sharpness_K_factor = 0x0000;

	/* UART handler */

	do {
		if(!process_data)
		{
			rx_size = UART_get_rx( &g_uart, &rx_byte, 1); //sizeof(g_rx_data) );
			if(rx_size > 0)
			{
				rx_ms_count = 10;
				rx_ms_count1 = 0;
				rx_tmr_done = 1;  // start/restart the timer
				g_rx_data[rx_cnt] = rx_byte;
				rx_cnt = rx_cnt + 1;
			}
		}
		else
		{
			crc16 = generateCRC((const uint8_t *)&g_rx_data[0], rx_cnt-2);
			rcvd_crc16 = ((g_rx_data[rx_cnt - 1] << 8) & 0xff00) | g_rx_data[rx_cnt - 2];
			UART_Handler();
			process_data = 0;
			rx_cnt = 0;
		}
	} while (1);


    return 0;
}

/**********************************************************/
/**********************FUNCTION CALLS**********************/
/**********************************************************/


void UART_Handler()
{
#if 1
	uint16_t crc16;
	uint16_t hue_value;


	switch(g_rx_data[0])
	{
	case 0x01:
	case 0x0a:

		for(i = 0; i < rx_cnt-2; i++)
			g_tx_data[i] = g_rx_data[i];

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], rx_cnt-2);

		g_tx_data[rx_cnt - 1] = crc16 >> 8;
		g_tx_data[rx_cnt - 2] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], rx_cnt);

		break;

	case 0x09:

		g_tx_data[0] = (uint8_t)0x09; //(temp);

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], 1);

		g_tx_data[2] = crc16 >> 8;
		g_tx_data[1] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], 3);

		break;

	case 0x02:	// Brightness & Contrast Adjustment

		*(volatile int*) Brightness_Gain_Address = g_rx_data[2];
		*(volatile int*) Contrast_Gain_Address = g_rx_data[4];

		g_tx_data[0] = (uint8_t)0x02;

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], 1);

		g_tx_data[2] = crc16 >> 8;
		g_tx_data[1] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], 3);

		break;

	case 0x3:  // Saturation & Hue Adjustment

		*(volatile int*) Saturation_Gain_Address = g_rx_data[2];
		hue_value = (g_rx_data[3] << 8) + g_rx_data[4];
		*(volatile int*) Hue_Gain_Address = hue_value;

		g_tx_data[0] = (uint8_t)0x03;

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], 1);

		g_tx_data[2] = crc16 >> 8;
		g_tx_data[1] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], 3);

		break;

	case 0x4:	// Sharpness Adjustment

		*(volatile int*) Sharpness_K_factor = (g_rx_data[1] << 8) + g_rx_data[2];

		g_tx_data[0] = (uint8_t)0x04;

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], 1);

		g_tx_data[2] = crc16 >> 8;
		g_tx_data[1] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], 3);

		break;

	case 0x5:	// Demo Selection

		if (g_rx_data[1] == 1u) {
			GPIO_set_output(&g_gpio_out, GPIO_4, 1u);
		} else {
			GPIO_set_output(&g_gpio_out, GPIO_4, 0u);
		}

		g_tx_data[0] = (uint8_t)0x05;

		crc16 = generateCRC((const uint8_t *)&g_tx_data[0], 1);

		g_tx_data[2] = crc16 >> 8;
		g_tx_data[1] = crc16 & 0xff;

		UART_send( &g_uart, &g_tx_data[0], 3);

		break;
	}

#endif

}



/* CRC algorithm */
uint16_t CRC16(uint16_t crc, uint8_t data)
{
	const uint16_t Poly16=0xA001;
	uint16_t LSB, i;
	crc = ((crc^data) | 0xFF00) & (crc | 0x00FF);
	for (i=0; i<8; i++) {
		LSB=(crc & 0x0001);
		crc=crc/2;
		if (LSB)
			crc=crc^Poly16;
	}
	return(crc);
}

/* Generate CRC */
uint16_t generateCRC(const uint8_t * data, int length)
{
	if(length < 1)
		return 0;
	uint16_t crc = 0xFFFF;
	for (int i = 0; i < length; i++) {
		crc = CRC16 (crc, data[i] );
	}

	return crc;
}

